	title "FELDHELL modem written by N1OWU with some ideas from KC7WW"
	page    132,60

;
; $Header: E:/HAM/EVM/RCS/feldhell.asm 1.4 1997/12/23 14:02:15 dbraun Exp $
;
; $Log: feldhell.asm $
; Revision 1.4  1997/12/23 14:02:15  dbraun
; Made RX bandwidth narrower, and renamed a few variables.
;
; Revision 1.3  1997/10/06 19:15:39  dbraun
; Added xmit filter and other nice stuff
;
; Revision 1.2  1997/10/06 10:57:50  dbraun
; Many changes and improvements
;
; Revision 1.1  1997/10/04 06:29:41  dbraun
; Initial revision
;
;


; This is a modem for Feldhell.  This uses CW (e.g. AM) modulation
; at a 122.5 Hz baud rate.
;
; The demodulator measures the signal amplitude 122.5 * 4, or 490, times
; per second, and each time sends a data byte (0 to 255) out the serial
; port at the standard 19200 baud rate.  For every 2nd transmitted
; byte (245 times per second),
; it expects to receive a command byte from the host program that tells
; it whether to transmit a mark tone, silence, or not to transmit.
; The advantage of this protocol is that the "standard" EVM serial
; port connections can be used, and no rewiring is necessary.
;
; A feature of the algorithm is the exact rate conversion from the
; 8000 Hz sampling rate to the 490 Hz output rate.
;
; Note that the host program's timing is derived from the rate of
; the received byte stream.
;
; The command "opcodes" sent from the host program are as follows:
;
; xxxxx110    Transmit a mark tone
; xxxxx101    Transmit a space tone (unused)
; xxxxx100    Transmit silence (space for Feldhell)
; xxxxx0xx    Do not transmit
;
; Note that some more opcodes should be implemented to e.g. reset the modem.
;
; There also ought to be a watchdog timer that turns of the transmitter
; if an opcode has not been received for a couple of seconds (e.g
; if the host program is very busy doing something else or has crashed).


	NOLIST
	include 'leonid.asm'
	LIST

	opt	mu


;******************************************************************************

UpLED   macro mode      ;UP line (a red LED connected)
        b\mode #1,X:$FFE4
        endm

DownLED macro mode      ;DOWN line (a red LED connected)
        b\mode #2,X:$FFE4
        endm

YellowLED macro mode    ;CAT line (a yellow LED connected)
        b\mode #3,X:$FFE4
        endm

DCDLed  macro mode      ; DCD (Green) LED clr/set/chg
        b\mode #13,X:$FFE4
        endm

;******************************************************************************

; Important constants
sampleFreq equ	8000.0
convOutRate equ	    800
convInRate  equ     49
outputRate  equ	sampleFreq*convInRate/convOutRate  ; should be 490

RxOutScale  equ	    6.0   ; Relationship between signal level (0.0 .. 1.0)
			  ; and data output level (0 .. 255)

RxGain      equ     0.0   ; CODEC input gain


markFreq   equ  1000.0
baudRate   equ	122.5

inputFilterSize   equ     48    ;Number of taps in input bandpass filter.

channelFilterBW	  equ     300   ; bandwidth in Hz
channelFilterSize equ     128   ; make this suitable for the chosen BW

outputFilterSize equ  1000     ;  Should be at least as large as outConvRate

convBufSize equ     outputFilterSize+convInRate+2  ; 2 for safety :-)

xmitFilterSize equ	48   ; bandpass filter for output signal

buflen	equ	4 	; CODEC buffer

maxTxCarAmpl  equ    0.5    ; Less than 1.0 to avoid overflow in filters

PTTRightChannel equ 1



      if PTTRightChannel
PTT     macro mode      ;PTT line: clr/set/chg
        b\mode #4,X:$FFE4
        endm
      else
PTT     macro mode      ;PTT line: clr/set/chg
        b\mode #0,X:$FFE4
        endm
      endif

;******************************************************************************
;************ MAIN PROGRAM ****************************************************

	org     P:user_code
main

        ori #%00000100,omr      ;enable the sine ROM table
                                ;this is for the IQ routine

;---------------------------------------------------------------------------
; Initialize pointer registers, r2, r7
; registers r0, r1, r4, and r5 are used as temporaries
;---------------------------------------------------------------------------

        move    #buffer+2,r7  ; r7 & m7 point to desired sample buffer
        move    #buflen*4-1,m7

        move    #buffer,r2
        move    #buflen*4-1,m2

;
; Program headphone,lineout enabled, speaker off, attn=0
;
	ctrlcd	1,r2,buflen,MIC,RxGain,RxGain,LINEO|HEADP,0.0,0.0
        opencd  sampleFreq/1000.0,HPF


;--------------- Main Loop -------------------------------------------------
main_loop 

	waitblk r2,buflen,1		; wait for one sample

	; Grab the RX data values
	
	move    X:(r2)+,a 	        ; Load the left channel input.
	;;move	a,X:<input_sig

	move    X:(r2),a     		; Load the right channel input.
	move	a,X:<input_sig

	; Transmitter part

	move X:<TxCarFreq,a   ;update the instant carrier frequency
	move X:<TxCarPhase,x0 ;get the carrier phase
	add x0,a              ;and advance it
	move a1,X:<TxCarPhase
	jsr <IQ               ;compute the sin/cos (modifies a,b,x,y,r0,m0,n0)


	; Multiply by the current TX amplitude 
	move	a,x0
	move	X:<TxCarAmpl,x1
	mpyr	x0,x1,a

	
	; Mute if not transmitting
	PTT tst
	.if <cc>
	  clr a
  	.endi


	; save in the xmit filter buffer
        move    #xmitFilterSize-1,m0
	move	X:<xmitFilterPtr,r0  
	nop
	move	a,X:(r0)+         ;put output into the filter tap
	move	r0,X:<xmitFilterPtr  

	; LPF it
        move    #xmitFilter,r4     ; R4 points to filter coefficients
	nop
	clr a	X:(r0)+,x0 Y:(r4)+,y0
	.loop	#xmitFilterSize-1
	  mac   x0,y0,a  X:(r0)+,x0  Y:(r4)+,y0
	.endl
	macr	x0,y0,a          ;a = filtered output

	move	a,Y:(r2)+        ;put to transmitter output (left channel)
  	move	a,X:<xmit_sig


	; Get the desired RX signal to actually process:
	; either the left (or right) input, or if we are are transmitting,
	; the tx signal (for loopback testing).

	PTT tst
	.if	<cs>
	  move	X:<xmit_sig,a
	  asr   a        ; scale its amplitude to be similar to rx sig
	.else
	  move	X:<input_sig,a
	.endi

	; stick it into the input filter buffer
	move	X:<inputFilterPtr,r0
        move    #inputFilterSize-1,m0
	nop
	move	a,X:(r0)+
	move	r0,X:<inputFilterPtr

;----------------------------------------------------------------------------
; Output RX + TX signal for monitoring
;----------------------------------------------------------------------------

	move    X:<xmit_sig,a	        ; Get current TX sig
	asr	a
	asr	a			; reduce its volume
	move	X:<filtered_sig,x0	; Get filtered RX signal
	add	x0,a			; mix them
	move    a,Y:(r2)+	        ; Put value in right channel (speaker).

	move	(r2)+			; R2 gets incremented 4 times


;---------------------------------------------------------------------------
; Perform input bandpass + I/Q filtering
; Input/Output sample rate = 8000 SPS 
;---------------------------------------------------------------------------

;---------------------------------------------------------------------------
;  Do I and Q channel filtering together in one loop
;---------------------------------------------------------------------------
        move    #inputFilterSize-1,m0
	move	X:<inputFilterPtr,r0	; R0 points to data

        move    #>-1,m4     		; linear addressing
        move    X:<curInputFilterI,r4  	; R4 points to I filter coefficients
        move    #>-1,m5     		; linear addressing
        move    X:<curInputFilterQ,r5  	; R5 points to Q filter coefficients
	nop

        clr     a                Y:(r4)+,y0     ; priming registers
        clr     b    X:(r0)+,x0  Y:(r5)+,y1
        .loop	#inputFilterSize-1			; allow interrupts
	mac     x0,y0,a              Y:(r4)+,y0
	mac     x0,y1,b  X:(r0)+,x0  Y:(r5)+,y1
	.endl
        macr    x0,y0,a				; final rounding
        macr    x0,y1,b				; I/Q values are in a/b

        move    #channelFilterSize-1,m0
	move	X:<channelFilterIPtr,r0
	nop
	move	a,X:(r0)+          ;save a in I buffer
	move	r0,X:<channelFilterIPtr

        move    #channelFilterSize-1,m0
	move	X:<channelFilterQPtr,r0
	nop
	move	b,X:(r0)+          ;save b in Q buffer
	move	r0,X:<channelFilterQPtr


;---------------------------------------------------------------------------
; Perform channel filtering
; Input/Output sample rate = 8000 SPS
; We do a total of 2 filtrations: mark I, mark Q
;---------------------------------------------------------------------------

        move    #channelFilterSize-1,m0
        move    #channelFilterSize-1,m1
        move    #>-1,m4     		; linear addressing

	move	X:<channelFilterIPtr,r0	; R0 points to I data
	move	X:<channelFilterQPtr,r1	; R1 points to Q data
        move    X:<curMarkFilter,r4     ; R4 points to mark filter coefficients
        nop

        clr     a    X:(r0)+,x0 
        clr     b    X:(r1)+,x1  Y:(r4)+,y0
        .loop	#channelFilterSize-1		; allow interrupts
        mac     x0,y0,a    X:(r0)+,x0 
        mac     x1,y0,b    X:(r1)+,x1  Y:(r4)+,y0
	.endl
        macr    x0,y0,a			; final rounding
        macr    x1,y0,b			; I/Q mark values are in a/b

        ; Now we have a filtered I/Q value in a/b
        ; Get the magnitude of the I/Q data: sqrt(I*I+Q*Q)

	; Save one of them for audio monitoring
	move	a,X:<filtered_sig

        jsr	IQmag
        move	b,a

        ; Adjust the gain a bit
        asl	a

        ; limit the value
	;(i.e. clip instead of wrapping around if it overflows)

        move	a,a1
        move	#0,a0

        ; Set/reset LEDs based on mark-space decision
        ; If the signal is sufficiently high, turn on the Mark LED.
        ; Otherwise turn on the space LED.

	UpLED clr
        DownLED clr
        move	#>0.5,x0
        cmp	x0,a
        .if <gt>
          UpLED	set
  	.else
          DownLED set
  	.endi

	; Do the convInRate/convOutRate rate conversion

	; Add the data value convInRate times to the conversion buffer
        
        move    #convBufSize-1,m0
	move	X:<convBufInPtr,r0
        .loop	#convInRate
          move     a,X:(r0)+
	.endl
	move	r0,X:<convBufInPtr


	; See if we have accumulated enough samples since the last 
	; output was generated

	move	X:<convBufOutPtr,r1

	; This tests if r0 is outputFilterSize words ahead of r1
	; r0 and r1 are left unchanged.

	; TODO: is the "+1" needed?
	move		r0,a
	move		r1,x0
	sub     x0,a	#>outputFilterSize+1,x0
	jmi     <_wrap
	cmp     x0,a				; wp - rp > threshold
	jlo     <notyet
	jmp     <_gotenough
_wrap	move		#>convBufSize,x1
	add     x1,a    ; buffer wraparound, rp - wp + length > threshold
	cmp     x0,a
	jlo     <notyet

_gotenough

	; If there are outputFilterSize samples, do output filtration over
	; them, generate an output value, and advance the conversion
	; output pointer

	move	X:<convBufOutPtr,r0
        move    #convBufSize-1,m0
        move    #outputFilter,r4     ; R4 points to filter coefficients
        nop

        clr     a    X:(r0)+,x0  Y:(r4)+,y0
        .loop	#outputFilterSize-1		; allow interrupts
        mac     x0,y0,a    X:(r0)+,x0  Y:(r4)+,y0
	.endl
        macr    x0,y0,a	     (r0)+	; Final rounding

	; Update the output pointer
	move	X:<convBufOutPtr,r0
	move	#convOutRate,n0
	nop
	move	(r0)+n0
	move	r0,X:<convBufOutPtr   

	btst #0,X:<spy
	.if <cs>
	  jsr <SpyA
  	.endi


	move a,y0             ;scale the output data
	move #>RxOutScale/32768.0,x0
	mpyr x0,y0,a #>$00,x0
	tlt x0,a
	move #>$FF,x0
	cmp x0,a
	tgt x0,a

	; Put byte (0..255) of data to serial port
	move	a1,x0

	btst #0,X:<spy
	.if <cc>
	  putc 
	.endi


	; Use a 1-bit counter to count every other data output
	bchg #0,X:<char_counter
	btst #0,X:<char_counter
	.if <cc>

	  ; Process any incoming opcode bytes from the serial port
	  btst #0,X:<spy
	  .if <cc>
	    lookc	0	; flush any input

	    .if <cc>	; there is a opcode byte
	      ; Test for things to do

	      btst	#2,x0	; TX/RX bit
	      .if <cs>
		PTT	set

		btst #1,x0	; TX-mark bit
		.if <cs>
		  move #maxTxCarAmpl,x0
		  move x0,X:<TxCarAmpl
		.else
		  move #0.0,x0
		  move x0,X:<TxCarAmpl
		.endi
	      .else
		PTT	clr
	      .endi
	    .endi
	  .endi
        .endi

notyet

	jmp	main_loop


PI      equ     3.14159265358979323846

;this routine computes a cosine/sine pair using the sine ROM
;with a second order (linear+quadrature) approximation between table points
IQ                              ;x0 = angle ( -1 = -PI, +1 = +PI)
        move #>$80,x1   ;shift out 8 most significant bits
        mpy x0,x1,a  #>$FF,x0
        move x0,m0
        and x0,a     #>$100,x0
        or x0,a      #<$40,n0
        move a1,r0      ;put the 8 most significant bits into r0 with offset = $100
        move a0,y0      ;save the remaining bits in y0
        jclr #23,y0,SinTable_lev2
          move (r0)+
SinTable_lev2
        move Y:(r0+n0),x0       ;x0 = coarse cosine
        move Y:(r0),x1          ;x1 = coarse sine
        mpyr x1,y0,a  #PI/256.0,y1
        tfr x0,a  a,x1
        macr -x1,y1,a           ;a = fine cosine
        mpyr x0,y0,b  Y:(r0),x1
        tfr x1,b  b,x1
        macr x1,y1,b  #PI*PI/2.0/65536.0,y1  ;b = fine sine
        mpyr y0,y0,a  a,x0
        move a,y0
        mpyr y0,y1,a
        tfr x0,a  a,y1
        macr -x0,y1,a  b,x1     ;a = super fine cosine
        macr -x1,y1,b           ;b = super fine sine
        rts                     ;x,y are modified
                                ;r0,m0,n0 are modified
                                ;maximum error is about 0.7E-6
                                ;execution time 4+64+4 clock cycles
                                ;including "jsr <IQ_lev2" and "rts"


; IQmag calculates the magnitude of a complex signal in a/b:
;     sqrt(I*I+Q*Q)

IQmag
	move	a,x0
	mpy	x0,x0,a
	move	b,x0
	mac	x0,x0,a  ; A = (I*I) + (Q*Q)  w/ 48-bit precision
	move    a0,y0
	move    a1,y1
	jsr	sqrt3    ; square root in B
	rts


; Full Precision Square Root by Polynomial Approximation. 
; Motorola DSP Operation, 6501 Wm. Cannon Drive W., Austin, Tx., 78735.
; Last Update 09 Feb 87   Version 1.0
;
; Full 23 bit precision square root routine using
; a successive approximation technique.
;
; y  = double precision (48 bit) positive input number
; b  = 24 bit output root
;
; a  = temporary storage
; x0 = guess
; x1 = bit being tested
; y1:y0 = input number
 
sqrt3
        clr     b    #<$40,x0           ; init root and guess
        move                    x0,x1   ; init bit to test
        do      #23,_endl
                                        ;START OF LOOP
        mpy     -x0,x0,a                ; square and negate the guess
        add     y,a                     ; compare to double precision input
        tge     x0,b                    ; update root if input >= guess
        tfr     x1,a                    ; get bit to test
        asr     a                               ; shift to next bit to test
        add     b,a             a,x1            ; form new guess
        move                    a,x0            ; save new guess
_endl                                    ;END OF LOOP
        rts



SpySync move a10,L:<SpySave     ;output: carry=1 => no spy request
        move a2,X:<SpySave+1    ;carry=0 => spy request !
        move x0,Y:<SpySave+1    ;512 words (jsr <SpyA) should follow
        move x1,Y:<SpyCount
        move X:<SpyCount,a
        tst a
        jne <SpyCont
        lookc 0
        jcs <Spy_end
        move #>'S',a
        cmp x0,a
        ori #$01,ccr
        jne <Spy_end
        move #>'P',x0
        putc
        move #>0,x0      ; send a 0 meaning non-complex data
        putc
        move #>512,a
        move a,X:<SpyCount
SpyCont andi #$FE,ccr
        jmp <Spy_end

SpyA    move a10,L:<SpySave
        move a2,X:<SpySave+1
        move x0,Y:<SpySave+1
        move x1,Y:<SpyCount
        move X:<SpyCount,a
        tst a
        jne <Spy_copy

Spy_check
        lookc 0
        jcs <Spy_end
        move #>'S',a
        cmp x0,a
        jne <Spy_end
        move #>'P',x0
        putc
        move #>0,x0      ; send a 0 meaning non-complex data
        putc
        move #>512,a
Spy_copy
        move #>1,x0
        sub x0,a
        move a,X:<SpyCount

        move X:<SpySave,a
        rep #8
          lsr a
        move a1,x0
	putc
        move X:<SpySave,a
        rep #16
          lsr a
        move a1,x0
        putc

Spy_end move L:<SpySave,a10
        move X:<SpySave+1,a2
        move Y:<SpySave+1,x0
        move Y:<SpyCount,x1
        ;rts
        rti		;N1OWU: the RTI instruction restores the SR


;*****************************************************************************
; Data definitions
;-----------------------------------------------------------------------------

;----------------------------------------------------------------------------
; INTERNAL RAM
;----------------------------------------------------------------------------

        LOMEM X:$0000,Y:$0000,L:$0000
        HIMEM X:$00FF,Y:$00FF,L:$00FF

        org L:user_data

SpySave dc 0,0
SpyCount dc 0
SpyTmp   ds 1

LastL	= *
	org     X:LastL
	org	Y:LastL

	org	X:

spy		dc	0	;runtime-patchable SPY switch
input_sig       ds      1       ;storage for (left or right) input signal
filtered_sig	ds	1	; current channel-filtered signal (for monitoring)
mark_sig 	ds	1
data_val	ds	1	; demodulated data value (-.999 to .999)
xmit_sig	ds	1	; most recent xmit signal value


; Pointers to the sets of filter taps currently
; in use.  

curInputFilterI dc	feldhellInputFilterI
curInputFilterQ dc	feldhellInputFilterQ
curMarkFilter	dc	feldhellMarkFilter

; These point to the current locations of the filters' data buffers.

inputFilterPtr	dc	inputFilterBuf
channelFilterIPtr dc	channelFilterIBuf
channelFilterQPtr dc	channelFilterQBuf
xmitFilterPtr	dc	xmitFilterBuf

; I/O pointers for the conversion and output filter buffer

convBufInPtr	dc	convBuf
convBufOutPtr	dc	convBuf

; The transmitter's NCO data.

TxCarFreq	dc	2.0*markFreq/sampleFreq
TxCarPhase	dc	0
TxCarAmpl	dc	0.0

char_counter		dc	0

; CODEC buffer
buffer	dsm	buflen*4	;CODEC buffer


; Data buffer for the input filter.
inputFilterBuf	dsm	inputFilterSize	; storage for input data  (circular buffer)


	org	Y:

	include 'firmacs.asm'	; various filter generation macros


; Input filter coefficients

fl = (markFreq-(4*baudRate))/sampleFreq
fh = (markFreq+(4*baudRate))/sampleFreq
feldhellInputFilterI    BlackmanHarrisFilterI	inputFilterSize,fl,fh
feldhellInputFilterQ    BlackmanHarrisFilterQ	inputFilterSize,fl,fh




;----------------------------------------------------------------------------
; EXTERNAL RAM
;----------------------------------------------------------------------------

        LOMEM X:$1000,Y:$1000,L:$1000
        HIMEM X:$37FF,Y:$37FF,L:$37FF

	org	L:$1000   ; Above last P memory location

LastL	= *

	org     X:LastL
	org	Y:LastL

	org	X:

channelFilterIBuf  dsm	channelFilterSize ; storage for channel filters (circ buffer)
channelFilterQBuf  dsm	channelFilterSize ; storage for channel filters (circ buffer)
convBuf 	   dsm  convBufSize  
xmitFilterBuf	   dsm	xmitFilterSize	


	org	Y:


fl = (markFreq-(0.5*channelFilterBW))/sampleFreq
fh = (markFreq+(0.5*channelFilterBW))/sampleFreq
feldhellMarkFilter    BasicBandpassFilter	channelFilterSize,fl,fh


;outputFilter   dup   outputFilterSize
;	dc	1.0/outputFilterSize   ; Filter simply averages the signal
;	endm
 

; Note: the filter size must be big enough to actually work.
; Since we only sample it every convOutRate samples, it must
; be at least that large, or sone data will be simply ignored...
; To avoid aliasing, the cutoff frequency must be less than half
; the effective sampling rate, outputRate.

fh = (0.8*outputRate)/(sampleFreq*convInRate)
outputFilter    BasicBandpassFilter	outputFilterSize,0,fh


fl = (markFreq-(2.0*baudRate))/(sampleFreq)
fh = (markFreq+(2.0*baudRate))/(sampleFreq)
xmitFilter    BasicBandpassFilter	xmitFilterSize,fl,fh

    end
