// evmhellDoc.cpp : implementation of the CEvmHellDoc class
//


#include "stdafx.h"
#include "evmhell.h"

#include "evmhellDoc.h"
#include "evmhellView.h"
#include "evmhellXmitMessages.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif



/////////////////////////////////////////////////////////////////////////////
// CEvmHellDoc

IMPLEMENT_DYNCREATE(CEvmHellDoc, CRichEditDoc)

BEGIN_MESSAGE_MAP(CEvmHellDoc, CRichEditDoc)
	//{{AFX_MSG_MAP(CEvmHellDoc)
	ON_COMMAND(ID_OPERATE_CLEARTXBUFFER, OnOperateCleartxbuffer)
	ON_COMMAND(ID_OPERATE_TRANSMIT, OnOperateTransmit)
	ON_COMMAND(ID_OPERATE_IMMEDIATERECEIVE, OnOperateImmediatereceive)
	ON_COMMAND(ID_OPERATE_RECEIVE, OnOperateReceive)
	ON_COMMAND(ID_OPERATE_DISPLAYOFF, OnOperateDisplayoff)
	ON_COMMAND(ID_OPERATE_DISPLAYON, OnOperateDisplayon)
  ON_UPDATE_COMMAND_UI(ID_OPERATE_DISPLAYOFF, OnUpdateOperateDisplayoff)
	ON_UPDATE_COMMAND_UI(ID_OPERATE_DISPLAYON, OnUpdateOperateDisplayon)
	ON_UPDATE_COMMAND_UI(ID_OPERATE_TRANSMIT, OnUpdateOperateTransmit)
	ON_COMMAND(ID_VIEW_BLACKONWHITE, OnViewBlackonwhite)
	ON_UPDATE_COMMAND_UI(ID_VIEW_BLACKONWHITE, OnUpdateViewBlackonwhite)
	ON_COMMAND(ID_VIEW_WHITEONBLACK, OnViewWhiteonblack)
	ON_UPDATE_COMMAND_UI(ID_VIEW_WHITEONBLACK, OnUpdateViewWhiteonblack)
	ON_COMMAND(ID_VIEW_CLEARDISPLAY, OnViewCleardisplay)
	ON_UPDATE_COMMAND_UI(ID_OPERATE_RECEIVE, OnUpdateOperateReceive)
	ON_COMMAND(ID_OPERATE_UPPERCASEXMIT, OnOperateUppercasexmit)
	ON_UPDATE_COMMAND_UI(ID_OPERATE_UPPERCASEXMIT, OnUpdateOperateUppercasexmit)
	ON_COMMAND(ID_OPERATE_FULLDUPLEX, OnOperateFullduplex)
	ON_UPDATE_COMMAND_UI(ID_OPERATE_FULLDUPLEX, OnUpdateOperateFullduplex)
  ON_COMMAND(ID_COMM_DUMMYPORT, OnCommDummyport)
	ON_UPDATE_COMMAND_UI(ID_COMM_DUMMYPORT, OnUpdateCommDummyport)
	ON_COMMAND(ID_COMM_COM1, OnCommCom1)
	ON_UPDATE_COMMAND_UI(ID_COMM_COM1, OnUpdateCommCom1)
	ON_COMMAND(ID_COMM_COM2, OnCommCom2)
	ON_UPDATE_COMMAND_UI(ID_COMM_COM2, OnUpdateCommCom2)
	ON_COMMAND(ID_COMM_COM3, OnCommCom3)
	ON_UPDATE_COMMAND_UI(ID_COMM_COM3, OnUpdateCommCom3)
	ON_COMMAND(ID_OPERATE_KICK, OnOperateKick)
	ON_COMMAND(ID_EDIT_XMITMESSAGES, OnEditXmitmessages)
	ON_COMMAND(ID_OPERATE_DIDDLE, OnOperateDiddle)
	ON_UPDATE_COMMAND_UI(ID_OPERATE_DIDDLE, OnUpdateOperateDiddle)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CEvmHellDoc construction/destruction

CEvmHellDoc::CEvmHellDoc()
{
	//  Add one-time construction code here
  //  DeleteContents() will be called before the Document is used.

  m_bRTF = FALSE;   // Do not save TX window as RTF
  nStrips = 0;
  curStrip = NULL;
  curStripIndex = 0;
  curStripRow = 0;
  curStripCol = 0;

  comPort = INVALID_HANDLE_VALUE;
  comPortSpeed = 19200;
  usingDummyPort = FALSE;

  mainThread = AfxGetThread();
  portThread = NULL;
  portThreadRunning = FALSE; 

}


CEvmHellDoc::~CEvmHellDoc()
{
   terminatePortThread();

  AfxGetApp()->WriteProfileString("Comm", "Port", comPortName);
  AfxGetApp()->WriteProfileInt("Tx", "UpperCase", upperCaseXmit);
  AfxGetApp()->WriteProfileInt("Tx", "ReverseVideo", reverseVideo);
  AfxGetApp()->WriteProfileInt("Tx", "FullDuplex", fullDuplex);
  AfxGetApp()->WriteProfileInt("Tx", "UseDiddle", useDiddle);
  shutdownSerialPort();

  
  // Dispose the old data strips
  int s;
  for(s=0; s<nStrips; s++) {
    DataStrip *sp = strips[s];
    delete sp;
  }
  strips.clear();
}



BOOL CEvmHellDoc::OnNewDocument()
{
 
  transmitting = FALSE;
  justStartedTransmitting = FALSE;
  idling = FALSE;
  stopWhenEmpty = FALSE;
  carrierFlag = FALSE;

  processRxData = TRUE;

  fullDuplex = AfxGetApp()->GetProfileInt("Tx", "FullDuplex", TRUE);
  upperCaseXmit = AfxGetApp()->GetProfileInt("Tx", "UpperCase", TRUE);
  reverseVideo = AfxGetApp()->GetProfileInt("Tx", "ReverseVideo", FALSE);
  useDiddle = AfxGetApp()->GetProfileInt("Tx", "UseDiddle", TRUE);


  // Open and load the font file

  txFontFileName = AfxGetApp()->GetProfileString("Tx", "FontFileName", "hellfont.hlf");

  HANDLE fontFile = CreateFile(txFontFileName, 
                    GENERIC_READ, 
                    FILE_SHARE_READ, 
                    0, 
                    OPEN_EXISTING,
                    0,
                    0);
  
  if (fontFile == INVALID_HANDLE_VALUE) {
    AfxMessageBox("Can't open font file " + txFontFileName);
    // then what???
  }

  ULONG nBytesRead;
  int bResult = ReadFile(fontFile, txCharFonts, sizeof(txCharFonts), &nBytesRead, NULL); 
  CloseHandle(fontFile);
  fontFile = INVALID_HANDLE_VALUE;



  // This calls CDocument::OnNewDocument(), which calls our override
  // of DeleteContents(). which initializes the data strips.
	if (!CRichEditDoc::OnNewDocument())
		return FALSE;

  // Initialize the serial port stuff,
  // This also deals with the portThread
  changeSerialPort(AfxGetApp()->GetProfileString("Comm", "Port", "DUMMY"));

  // Initialize the canned Xmit messages
  char msgName[80];

  for (int j=0; j<NUM_XMIT_MESSAGES; j++) {
    sprintf(msgName, "Message%d", j);
    xmitMessageTexts[j] = AfxGetApp()->GetProfileString("XmitMessages", msgName, "");
  }

	return TRUE;

}


void CEvmHellDoc::OnCloseDocument() 
{

//  terminatePortThread();

//  WriteProfileString("Comm", "Port", comPortName);
//  shutdownSerialPort();

  // Save the canned Xmit messages
  char msgName[80];

  for (int j=0; j<NUM_XMIT_MESSAGES; j++) {
    sprintf(msgName, "Message%d", j);
    AfxGetApp()->WriteProfileString("XmitMessages", msgName, xmitMessageTexts[j]);
  }


  CRichEditDoc::OnCloseDocument();   // This does a "delete this"...
}



void CEvmHellDoc::DeleteContents() 
{
 
  // Dispose the old data strips
  int s;
  for(s=0; s<nStrips; s++) {
    DataStrip *sp = strips[s];
    delete sp;
  }
  strips.clear();

  // Initialize one new data strip
  curStrip = new DataStrip;
  nStrips = 1;
  curStripIndex = 0;
  strips.push_back(curStrip);
  curStripRow = 0;
  curStripCol = 0;
  rxCount = 0;

  CDocument::DeleteContents();


  
  // Notify the view that the data size has changed

  CEvmHellRxView* rxView = getRxView();
  if (rxView != NULL)
    rxView->ScrollToEnd();
 
  UpdateAllViews(NULL);
}



CEvmHellRxView* CEvmHellDoc::getRxView()
{
  
  POSITION pos = GetFirstViewPosition();
  while (pos != NULL)
  {
    CView* pView = GetNextView(pos);
    if (pView != GetView()) {
      CEvmHellRxView* rxView = (CEvmHellRxView*)pView;
      ASSERT(rxView->IsKindOf( RUNTIME_CLASS( CEvmHellRxView )));
      return rxView;
    }
  }
  
  return(NULL); 
}



/////////////////////////////////////////////////////////////////////////////
// CEvmHellDoc serialization

void CEvmHellDoc::Serialize(CArchive& ar)
{
  if (ar.IsStoring())
  {
    // Dump out all the bytes in the order they were received, with no special
    // formatting.
    for (int s=0; s<nStrips; s++) {
      DataStrip *sp = strips[s];
      for (int col=0; col<COLS_PER_STRIP; col++) {
        for (int row=0; row< RX_CHAR_HEIGHT;row++) {
          if (sp == curStrip && row == curStripRow && col == curStripCol)
            goto finished;  // We're done
          ar << sp->pixels[col][row];
        }
      }
    }
finished: ;
          
  }
  else
  {
    // Read bytes and stuff them into strips, until there are no more bytes
    
    BYTE dataVal;
    for (;;) {
      
      try
      {
        ar >> dataVal;
      }
      catch (CFileException* e)
      {
        if (e->m_cause == CFileException::endOfFile)
          break;
        else
          throw;  // Pass it up
      }
      catch (CArchiveException* e)
      {
        if (e->m_cause == CArchiveException::endOfFile)
          break;
        else
          throw;  // Pass it up
      }
      
      processRxChar(dataVal, FALSE);
    }

    // Initialize the position of the scrolling
    getRxView()->ScrollToEnd();
  
    UpdateAllViews(NULL);
    
  }
}

/////////////////////////////////////////////////////////////////////////////
// CEvmHellDoc diagnostics

#ifdef _DEBUG
void CEvmHellDoc::AssertValid() const
{
	CRichEditDoc::AssertValid();
}

void CEvmHellDoc::Dump(CDumpContext& dc) const
{
	CRichEditDoc::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CEvmHellDoc commands


CRichEditCntrItem *CEvmHellDoc::CreateClientItem(REOBJECT *preo) const 
{
  return CRichEditDoc::CreateClientItem(preo);
}



void CEvmHellDoc::OnOperateCleartxbuffer() 
{
  // Tell the TxView to deal with it
  CEvmHellTxView* txViewPtr = (CEvmHellTxView*)GetView();
  txViewPtr->clearTxBuffer();
}



// This function gets called "frequently" in response to
// timer messages or periodic messages from the comPort thread.

void CEvmHellDoc::serviceDataQueues()
{
  static bool servicing = FALSE;

  if (servicing)
    return;  // Avoid infinite recursion, since this can call onIdle(), and onIdle
             // might possibly call this...

  servicing = TRUE;

  // Service the rxDataQueue.  Release the lock after getting each char
  // and before processing it, to avoid starving the comPort thread.

  rxDataQueueLock.Lock();
  
  while (!rxDataQueue.IsEmpty()) {
    int rxData = rxDataQueue.RemoveHead();
    rxDataQueueLock.Unlock();
    if (processRxData && (!transmitting || fullDuplex)) {
      processRxChar(rxData, TRUE);
    }
    rxDataQueueLock.Lock();
  }

  rxDataQueueLock.Unlock();

  // If the txData queue is not too full, shove some more data into it.
  // If there is a character ready to transmit, send all of its data.
  // Otherwise, send one column's worth of dummy data.
  
  txDataQueueLock.Lock();
  
  if (txDataQueue.GetCount() < 2*TX_CHAR_HEIGHT) {
    
    int txChar;
    bool isMorse;
    bool xmitIt;
    getNextTxChar(txChar, xmitIt, isMorse);

    // If we are idling or receiving, send only one column of blanks.
    // This causes the idling periods to be in units of one
    // column instead of one character.

    if (!xmitIt || txChar == NO_CHAR) {

      // Send a column of blanks
      for(int row=0;row<TX_CHAR_HEIGHT;row++) {
        int dataChar;
        
        dataChar = xmitIt ? MODEM_TX : MODEM_RX;
        txDataQueue.AddTail(dataChar);
      }
      
    } else if (!isMorse) {

      // Send a full, real Feldhell character
      int ncols = getTxFontWidth(txChar);

      for(int col=0;col<ncols;col++) {
        for(int row=0;row<TX_CHAR_HEIGHT;row++) {
          int dataChar;
                   
          dataChar = MODEM_TX;
          if (getTxFontBit(txChar, col, row))
            dataChar |= MODEM_MARK;
          txDataQueue.AddTail(dataChar);
        }
      }

    } else {

      // Send a Morse character
      CString dotsAndDashes;
      getMorseString(txChar, dotsAndDashes);

      // Send 3/2 dots' worth of silence first
      for (int k=0; k<(3*MORSE_DOT_LENGTH/2); k++) {
        txDataQueue.AddTail(MODEM_TX);
      }

      // Send each dot or dash, each followed by one dot of silence.
      for(int j=0; j<dotsAndDashes.GetLength(); j++) {
        int curDotDash = dotsAndDashes[j];
        int numDataBytes;

        if (curDotDash == '.')
          numDataBytes = MORSE_DOT_LENGTH;
        else if (curDotDash == '-')
          numDataBytes = 3*MORSE_DOT_LENGTH;
        else
          numDataBytes = 0;  // A mistake in the morse string format...

        // Send the dot or dash
        for (k=0; k<numDataBytes; k++) {
          txDataQueue.AddTail(MODEM_TX | MODEM_MARK);
        }

        // Send the one dot's worth of silence after the actual dot/dash.
        for (k=0; k<MORSE_DOT_LENGTH; k++) {
          txDataQueue.AddTail(MODEM_TX);
        }
      }

      // Send 1/2 dot's worth of silence after all the dots and dashes.
      // The total inter-char gap = 3/2 + 1 + 1/2 = 3
      for (k=0; k<(MORSE_DOT_LENGTH/2); k++) {
        txDataQueue.AddTail(MODEM_TX);
      }

      // Special case for " ": one more dot's worth of silence to make
      // inter-word spaces 7 dots long: 3 from the letters on either side of the " ",
      // and 4 from the " " itself.

      if (txChar == ' ') {
        for (k=0; k<(MORSE_DOT_LENGTH); k++) {
          txDataQueue.AddTail(MODEM_TX);
        }
      }

    }
  }
  txDataQueueLock.Unlock();

  servicing = FALSE;
}



void CEvmHellDoc::processRxChar(int rxChar, bool updateScreen)
{
  // This takes a data byte (from the serial port or elsewhere),
  // adds it to the data structures,
  // and possibly updates the display.
  
  
  rxCount ++;
  SetModifiedFlag();
  
  // Put into the current DataStrip
  curStrip->pixels[curStripCol][curStripRow++] = rxChar;
  if (curStripRow >= RX_CHAR_HEIGHT) {
    
    // If we just finished filling a column, draw it, if desired
    if (updateScreen) {
#if 0
      
      DrawingHint hint;
      hint.stripIndex = curStripIndex;
      hint.stripCol = curStripCol;
      UpdateAllViews(GetView(), 0L, &hint);  // Don't attempt to update the TxView
      
#else
      
      getRxView()->DrawNewStripData(curStripIndex, curStripCol);
    
#endif
    }
    
    // Point to the next column
    curStripRow = 0;
    curStripCol++;
    
    // If necessary, start a new strip
    if (curStripCol >= COLS_PER_STRIP) {
      // Add a newStrip
      curStrip = new DataStrip;
      nStrips++;
      curStripIndex++;
      strips.push_back(curStrip);
      curStripRow = 0;
      curStripCol = 0;
      
      if (updateScreen) {
        getRxView()->ResyncScrollSizes();
        // Tell the Rx view to scroll if necessary.
        getRxView()->ScrollForNewStrip();
      }
    
    }
  }
  
}



// This returns the number of columns needed to hold the given character.
// We can in principle support variable-width fonts...
int CEvmHellDoc::getTxFontWidth(int txChar)
{
  return CHAR_WIDTH;
  //return TX_FONT_WIDTH;
}



bool CEvmHellDoc::getTxFontBit(int txChar, int txCharCol, int txCharRow)
{

  // Careful not to overflow array
  ASSERT(txChar >= 0 && txChar < 256);

  // Work around mistakes in the font file.  (The rich edit control returns
  // a paragraph break as a null char instead of \n.)
  if (txChar == '\0' || txChar == '\n' || txChar == '\r' || txChar == '\t')
    txChar = ' ';


  // This is legal; if the font is smaller then the Tx size, we pad it
  if (txCharCol >= TX_FONT_WIDTH || txCharRow >= TX_FONT_HEIGHT)
    return 0;

  // The addressing of the array is determined by the format of the font file.
  return (txCharFonts[txChar][txCharCol][txCharRow] != 0);
}



void CEvmHellDoc::getMorseString(int txChar, CString& morseStr) {

  txChar = toupper(txChar);

  // Crude but effective...
  
  if (txChar == 'A')
    morseStr = CString(".-");
  else if (txChar == 'B')
    morseStr = CString("-...");
  else if (txChar == 'C')
    morseStr = CString("-.-.");
  else if (txChar == 'D')
    morseStr = CString("-..");
  else if (txChar == 'E')
    morseStr = CString(".");
  else if (txChar == 'F')
    morseStr = CString("..-.");
  else if (txChar == 'G')
    morseStr = CString("--.");
  else if (txChar == 'H')
    morseStr = CString("....");
  else if (txChar == 'I')
    morseStr = CString("..");
  else if (txChar == 'J')
    morseStr = CString(".---");
  else if (txChar == 'K')
    morseStr = CString("-.-");
  else if (txChar == 'L')
    morseStr = CString(".-..");
  else if (txChar == 'M')
    morseStr = CString("--");
  else if (txChar == 'N')
    morseStr = CString("-.");
  else if (txChar == 'O')
    morseStr = CString("---");
  else if (txChar == 'P')
    morseStr = CString(".--.");
  else if (txChar == 'Q')
    morseStr = CString("--.-");
  else if (txChar == 'R')
    morseStr = CString(".-.");
  else if (txChar == 'S')
    morseStr = CString("...");
  else if (txChar == 'T')
    morseStr = CString("-");
  else if (txChar == 'U')
    morseStr = CString("..-");
  else if (txChar == 'V')
    morseStr = CString("...-");
  else if (txChar == 'W')
    morseStr = CString(".--");
  else if (txChar == 'X')
    morseStr = CString("-..-");
  else if (txChar == 'Y')
    morseStr = CString("-.--");
  else if (txChar == 'Z')
    morseStr = CString("--..");

  else if (txChar == '0')
    morseStr = CString("-----");
  else if (txChar == '1')
    morseStr = CString(".----");
  else if (txChar == '2')
    morseStr = CString("..---");
  else if (txChar == '3')
    morseStr = CString("...--");
  else if (txChar == '4')
    morseStr = CString("....-");
  else if (txChar == '5')
    morseStr = CString(".....");
  else if (txChar == '6')
    morseStr = CString("-....");
  else if (txChar == '7')
    morseStr = CString("--...");
  else if (txChar == '8')
    morseStr = CString("---..");
  else if (txChar == '9')
    morseStr = CString("----.");

  else if (txChar == '/')
    morseStr = CString("-..-.");
  else if (txChar == '?')
    morseStr = CString("..--..");
  else if (txChar == '+')
    morseStr = CString(".-.-.");   // AR code
  else if (txChar == '.')
    morseStr = CString(".-.-.-");
  else if (txChar == ',')
    morseStr = CString("--..--");
  else if (txChar == '-')
    morseStr = CString("-....-");
  else if (txChar == ',')
    morseStr = CString("--..--");
  else if (txChar == '*')
    morseStr = CString("........");  // Error symbol HH

  else
    morseStr = CString("");


}



BOOL CEvmHellDoc::SaveModified() 
{
	return TRUE;   // Say it is always OK to close the document
}


char * CEvmHellDoc::getTxRxStatusStr()
{
  
  if (transmitting) {
    //ASSERT(!(stopWhenEmpty && idling));  //These can both be on for a brief time

    if (idling)
      return "Idling";
    else if (stopWhenEmpty)
      return "Finishing Transmission";
    else 
      return "Transmitting";
  } else {

    ASSERT(!stopWhenEmpty && !idling && !justStartedTransmitting);

    return "Receiving";
  }
}



void CEvmHellDoc::getNextTxChar(int& txChar, bool& xmitIt, bool& morse)
{
  // If we are transmitting, this finds the next character to transmit,
  // puts it into txChar, and returns TRUE.
  // If there is nothing to send at the moment, it will set TxChar to
  // NO_CHAR.  If we are receiving it returns FALSE, and txChar is NO_CHAR.
 
  // It updates the various Tx status flags;
  // It also returns a blank char as a txDelay/txTail filler as required.

  // If the char is intended to be sent as Morse instead of Feldhell,
  // morse is set.
 
  CEvmHellTxView* txViewPtr = (CEvmHellTxView*)GetView();
  
  morse = FALSE;
  xmitIt = FALSE;  // Default condition when receiving
  txChar = NO_CHAR; 

  if (transmitting) {
    int nextChar;

    xmitIt = TRUE;

    if (justStartedTransmitting) {
      nextChar = ' ';    //txDelay filler char
      justStartedTransmitting = FALSE;
    } else {
      txViewPtr->getNextXmitChar(nextChar, morse);
      if (upperCaseXmit && nextChar != NO_CHAR) {
        nextChar = toupper(nextChar);
      }
    }

    if (nextChar != NO_CHAR) {
      // The buffer was not empty
      idling = FALSE;

      txChar = nextChar;
 
    } else {
      // The buffer is empty:
      if (stopWhenEmpty) {
        // Actually shut off transmitter
        transmitting = FALSE;
        stopWhenEmpty = FALSE;
        idling = FALSE;

        txChar = ' ';  // txTail filler char

      } else {
        idling = TRUE;

        if (useDiddle)
          txChar = DIDDLE_CHAR;  // A special tiny dot on the font.
        else
          txChar = ' ';          // Blank
      }

      // We are spontaneously changing the TxRx state (i.e. not
      // based on any UI input events.  We must force Windows to
      // do its event stuff to update the status indicators, etc.
      AfxGetApp()->OnIdle(0);
    }

  }

}



bool CEvmHellDoc::activatePortThread(unsigned int priority)
{
  ASSERT(portThread == NULL);
  
  portThread = AfxBeginThread(portThreadControllingProc, this, priority);
 
  ASSERT(portThread != NULL);
  return(TRUE);
}


// This is passed to AfxBeginThread().
// That's why it's static.
// Called only from activatePortThread()!!!
UINT CEvmHellDoc::portThreadControllingProc(LPVOID param)
{
  CEvmHellDoc* myself = (CEvmHellDoc*) param;
  ASSERT(myself != NULL);
  ASSERT(myself->IsKindOf( RUNTIME_CLASS( CEvmHellDoc )));

  return myself->portThreadMainLoop();
}


// Called only from portThreadControllingProc()!!!

UINT CEvmHellDoc::portThreadMainLoop()
{
  
  portThreadRunning = TRUE; 
  terminatePortThreadFlag = FALSE;
  int dummyDataVal = 0;
  int messageCounter = 0;   // For keeping track of when to send messages
  int txCounter = 0;       // For keeping track of when to sent Tx data bytes
  int lastDataByte = 0;
  
  ASSERT(comPort != INVALID_HANDLE_VALUE || usingDummyPort);
  
  for (;;) {
    
    int dataVal;
    
    if (usingDummyPort) {
      // Return dummy data at approximately the right rate
      SleepEx(1,FALSE);  // Should be 2.04082 mS
      dataVal = dummyDataVal;
    } else {
      
      // Actually read the port
      ULONG nBytesRead;
      unsigned char inputBuf[80];
      
      ASSERT(comPort != INVALID_HANDLE_VALUE);
      int bResult = ReadFile(comPort, inputBuf, 1, &nBytesRead, NULL); 
      
      // Note that we set up the port to time out if there is no data
      // after a while.  This makes it possible to terminate this thread
      // if no data is coming in.

      // If the read times out due to the lack of incoming data, we  will
      // end up effectively getting a very slow data stream of zeros...
      
      if (bResult && nBytesRead == 1)
        dataVal = inputBuf[0] & 0xff;
      else
        dataVal = 0;
    }
    
    // Put the received data byte into the rxDataQueue.
    // If it is already full, toss out the oldest entry
    // to make room for the new.
 
    rxDataQueueLock.Lock();
    if (rxDataQueue.GetCount () >= MAX_RX_QUEUE_SIZE) {
      rxDataQueue.RemoveHead();   // Toss out oldest
    }
    rxDataQueue.AddTail(dataVal);
    rxDataQueueLock.Unlock();


    // For every Nth received character, send a character,
    // where "N" is the Rx/Tx pixel ratio.

    txCounter++;
    if (txCounter >= RX_TX_PIXEL_RATIO) {
      txCounter = 0;
      
      // Grab a Tx data byte from the txDataQueue and
      // send it out the serial port.  If the queue is empty,
      // stuff it with one column's worth of data, based on the 
      // last valid data byte, to maintain row synchronization.
      
      int txDataByte;

      txDataQueueLock.Lock();

      if (txDataQueue.IsEmpty()) {
        // If we have been transmitting, stuff a column of "TX silence",
        // otherwise stuff "RX".
        int filler = lastDataByte & MODEM_TX;
        for(int j=0;j<TX_CHAR_HEIGHT;j++)
          txDataQueue.AddTail(filler);
      }

      ASSERT(!txDataQueue.IsEmpty());
      txDataByte = txDataQueue.RemoveHead();
      txDataQueueLock.Unlock();
      
      // Actually send it out the serial (or dummy) port.
      
      if (usingDummyPort) {
        // Feed back tramsmitted data to be received data (in correct format)
        dummyDataVal = (txDataByte & MODEM_MARK) ?  200 : 50;
      } else if (comPort != INVALID_HANDLE_VALUE) {
        ULONG nBytesWritten;
        char outputBuf[1];
 
        outputBuf[0] = txDataByte;
        int bResult = WriteFile(comPort, outputBuf, 1, &nBytesWritten, NULL);
      }

      lastDataByte = txDataByte;  // Save in case queue empties
    }


    // It's much better to use a system timer instead...
#if 0

    // For every Nth received data byte, send a message to
    // the main thread to make it do its stuff.
    
    messageCounter++;
    if (messageCounter >= 10) {
      messageCounter = 0;
      ASSERT(mainThread != NULL);
      mainThread->PostThreadMessage(WM_RX_CHAR_MESSAGE, 0, 0);
    }
#endif

     
    if (terminatePortThreadFlag)  // set by main thread
      break;
  }
  
  portThreadRunning = FALSE; 
  
  return 0;
}


void CEvmHellDoc::terminatePortThread()
{
  // This is tricky, because the thread might be waiting for
  // data from the serial port...  
  // It works because the port is set to timeout after a while during reads.

  if (!portThread) {
    ASSERT(portThreadRunning == FALSE);
    return;
  }
  
  terminatePortThreadFlag = TRUE;   // Activate signal for the thread.

  do {
    Sleep(0);  // Give the listener thread a chance to run, see the flag, and exit.
  } while (portThreadRunning);

  terminatePortThreadFlag = FALSE;
  portThread = NULL;
}





bool CEvmHellDoc::changeSerialPort(CString newPortName)
{
  terminatePortThread();
  shutdownSerialPort();

  bool ok;

  if (newPortName == "DUMMY") {
    ok = setupDummySerialPort();
  } else {
    ok = setupSerialPort(newPortName);
  }

  // Start the thread that listens to the serial port.
  // The listener thread will send messages to our main thread,
  // which will relay them to the document.  
  
  if (ok) {
    ASSERT(comPort != INVALID_HANDLE_VALUE || usingDummyPort); 
    activatePortThread(THREAD_PRIORITY_TIME_CRITICAL);
  }

  return ok;

}


bool CEvmHellDoc::setupSerialPort(CString name)
{
  
  if (usingDummyPort || comPort != INVALID_HANDLE_VALUE) {
    return (false);  // Already opened
  }

  comPort = CreateFile(name, 
                    GENERIC_READ | GENERIC_WRITE, 
                    FILE_SHARE_READ | FILE_SHARE_WRITE, 
                    0, 
                    OPEN_EXISTING,
                    0,
                    0);
  
  if (comPort == INVALID_HANDLE_VALUE) {
    return FALSE;
  }

  comPortName = name;
  usingDummyPort = FALSE;

  // Set the port parameters

  DCB dcb;
  BuildCommDCB("baud=19200 parity=N data=8 stop=1", &dcb);
  dcb.fParity = FALSE;   // No parity checking
  dcb.fBinary = TRUE;
  dcb.fNull = FALSE;  // Do not strip 00 (null) chars
  dcb.fAbortOnError = FALSE;   // We don't to know about problems...
  dcb.fOutxCtsFlow = FALSE;    // Ignore CTS
  dcb.fOutxDsrFlow = FALSE;   // Ignore DSR
  dcb.fDsrSensitivity = FALSE;   // Ignore DSR
  dcb.fInX = FALSE;   // No Xon-Xoff business...
  dcb.fOutX = FALSE;  
  dcb.fDtrControl = DTR_CONTROL_ENABLE;  // DTR always on
  dcb.fRtsControl = RTS_CONTROL_ENABLE;   // RTS always on
  SetCommState(comPort, &dcb);

  // Set the port timeouts, so that the listener thread will
  // not wait forever if no data is available.

  COMMTIMEOUTS timeouts;
  timeouts.ReadIntervalTimeout = MAXDWORD; 
  timeouts.ReadTotalTimeoutMultiplier = MAXDWORD; 
  timeouts.ReadTotalTimeoutConstant = 200;   //  200 mS timeout value 
  timeouts.WriteTotalTimeoutMultiplier = 0;
  timeouts.WriteTotalTimeoutConstant = 0; 
  SetCommTimeouts(comPort, &timeouts);



  return TRUE;
}


bool CEvmHellDoc::setupDummySerialPort()
{
 
  if (usingDummyPort || comPort != INVALID_HANDLE_VALUE) {
    return (false);  // Already opened
  }

  comPortName = "DUMMY";
  usingDummyPort = TRUE;

  return TRUE;
}



void CEvmHellDoc::shutdownSerialPort()
{

  if (comPort != INVALID_HANDLE_VALUE)
    CloseHandle(comPort);

  comPort = INVALID_HANDLE_VALUE;
  usingDummyPort = FALSE;
  comPortName = "";
}


// This is not used by the "normal" data processing, only
// for exceptional cases.

void CEvmHellDoc::sendTxData(int dataByte)
{
  ULONG nBytesWritten;
  char outputBuf[1];
  
  if (usingDummyPort) {
    return;
  }

  if (comPort != INVALID_HANDLE_VALUE) {
    outputBuf[0] = dataByte;
    int bResult = WriteFile(comPort, outputBuf, 1, &nBytesWritten, NULL);
  }
}




/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// CEvmHellDoc commands



void CEvmHellDoc::OnOperateTransmit() 
{

  if (!transmitting) {
    transmitting = TRUE;
    justStartedTransmitting = TRUE;
    stopWhenEmpty = FALSE;
  } else {
    stopWhenEmpty = FALSE;
  }
}
	
void CEvmHellDoc::OnOperateImmediatereceive() 
{
  transmitting = FALSE;
  stopWhenEmpty = FALSE;
  idling = FALSE;
  justStartedTransmitting = FALSE;
}

void CEvmHellDoc::OnOperateReceive() 
{
	// If the buffer is already empty, the next attenpt to send
  // a character will shut off the transmitter.
  if (transmitting) {
    stopWhenEmpty = TRUE;
  }
}





void CEvmHellDoc::OnCommDummyport() 
{
	  changeSerialPort("DUMMY");
}

void CEvmHellDoc::OnUpdateCommDummyport(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(comPortName == "DUMMY");		
}



void CEvmHellDoc::OnCommCom1() 
{
  if (!changeSerialPort("COM1")) {
    AfxMessageBox("Can't open port COM1");
  }
}

void CEvmHellDoc::OnUpdateCommCom1(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(comPortName == "COM1");	
}



void CEvmHellDoc::OnCommCom2() 
{
  if (!changeSerialPort("COM2")) {
    AfxMessageBox("Can't open port COM2");
  }
}

void CEvmHellDoc::OnUpdateCommCom2(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(comPortName == "COM2");	
}



void CEvmHellDoc::OnCommCom3() 
{
  if (!changeSerialPort("COM3")) {
    AfxMessageBox("Can't open port COM3");
  }
}

void CEvmHellDoc::OnUpdateCommCom3(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(comPortName == "COM3");	
}






void CEvmHellDoc::OnOperateDisplayoff() 
{
	// Stop displaying received data
  processRxData = FALSE;
	
}

void CEvmHellDoc::OnOperateDisplayon() 
{
  // Resume displaying received data
  processRxData = TRUE;
}

void CEvmHellDoc::OnUpdateOperateDisplayoff(CCmdUI* pCmdUI) 
{
  pCmdUI->SetRadio(processRxData == FALSE);	
}

void CEvmHellDoc::OnUpdateOperateDisplayon(CCmdUI* pCmdUI) 
{
  pCmdUI->SetRadio(processRxData == TRUE);	
 
}

void CEvmHellDoc::OnUpdateOperateTransmit(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(transmitting == TRUE);
	
}

void CEvmHellDoc::OnUpdateOperateReceive(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(transmitting == FALSE || stopWhenEmpty  == TRUE);
}

void CEvmHellDoc::OnViewBlackonwhite() 
{
  reverseVideo = TRUE;
  UpdateAllViews(NULL);
	
}

void CEvmHellDoc::OnUpdateViewBlackonwhite(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(reverseVideo);
}

void CEvmHellDoc::OnViewWhiteonblack() 
{
	reverseVideo = FALSE;
  UpdateAllViews(NULL);
	
}

void CEvmHellDoc::OnUpdateViewWhiteonblack(CCmdUI* pCmdUI) 
{
  pCmdUI->SetRadio(!reverseVideo);
  
}


void CEvmHellDoc::OnViewCleardisplay() 
{
  // Erase all stored data.
  // Dispose the old data strips
 
  int s;
  for(s=0; s<nStrips; s++) {
    DataStrip *sp = strips[s];
    delete sp;
  }

  strips.clear();

  curStrip = NULL;
  nStrips = 0;
  curStripIndex = 0;
  curStripRow = 0;
  curStripCol = 0;

  // Initialize the data strips
  curStrip = new DataStrip;
  nStrips = 1;
  curStripIndex = 0;
  strips.push_back(curStrip);
  curStripRow = 0;
  curStripCol = 0;

  UpdateAllViews(NULL);
  
}


void CEvmHellDoc::OnOperateUppercasexmit() 
{
  upperCaseXmit = !upperCaseXmit;	
}

void CEvmHellDoc::OnUpdateOperateUppercasexmit(CCmdUI* pCmdUI) 
{
  pCmdUI->SetCheck(upperCaseXmit);
}

void CEvmHellDoc::OnOperateFullduplex() 
{
  fullDuplex = !fullDuplex;
}

void CEvmHellDoc::OnUpdateOperateFullduplex(CCmdUI* pCmdUI) 
{
  pCmdUI->SetCheck(fullDuplex);
}



BOOL CEvmHellDoc::OnSaveDocument(LPCTSTR lpszPathName) 
{
  // This saves the Rx data, not the Tx data.

  // We have to un-override our base classes' implementation
  // of this...

  // This will call our override of Serialize().
  return CDocument::OnSaveDocument(lpszPathName); 

}

BOOL CEvmHellDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
  // This loads the Rx Data, not the Tx data.

  // We have to un-override our base classes' implementation
  // of this...

  // This will call our override of Serialize().
  return CDocument::OnOpenDocument(lpszPathName); 


}


// This is mainly for debugging.
void CEvmHellDoc::OnOperateKick() 
{
  serviceDataQueues();	
}


// Put up the Property Sheet with all the canned Xmit messages.
void CEvmHellDoc::OnEditXmitmessages() 
{

 CEvmHellXmitMessageSheet msgSheet("Transmit Messages", NULL, NUM_XMIT_MESSAGES);

 //  Stuff the property sheet with the existing messages
 for(int j=0; j<NUM_XMIT_MESSAGES; j++) {
   msgSheet.messages[j].messageText = xmitMessageTexts[j];
 }

 if (msgSheet.DoModal() == IDOK) {

   //  Get back the messages
   for(int j=0; j<NUM_XMIT_MESSAGES; j++) {
     xmitMessageTexts[j] = msgSheet.messages[j].messageText;
   }
 }

}


void CEvmHellDoc::OnOperateDiddle() 
{
	useDiddle = !useDiddle;
}

void CEvmHellDoc::OnUpdateOperateDiddle(CCmdUI* pCmdUI) 
{
  pCmdUI->SetCheck(useDiddle);
}
