// evmhellDoc.h : interface of the CEvmHellDoc class
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_EVMHELLDOC_H__8A0E0ECF_366C_11D1_BB7C_0080489E8DF7__INCLUDED_)
#define AFX_EVMHELLDOC_H__8A0E0ECF_366C_11D1_BB7C_0080489E8DF7__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include <afxmt.h>   // For CCriticalSection
#include <vector>  // STL
//#include <queue>
#include <afxtempl.h>

using namespace std;



#define CHAR_WIDTH 7   // For both Rx and Tx (probably either 6 or 7)
#define RX_CHAR_HEIGHT  28  // We are technically doing 4x oversampling
#define TX_CHAR_HEIGHT  14  // Transmit w/ 2x oversampling

#define TX_FONT_WIDTH   6   // Can be smaller than CHAR_WIDTH
#define TX_FONT_HEIGHT  14   // Can be smaller than TX_CHAR_HEIGHT


#define CHARS_PER_STRIP  35
#define COLS_PER_STRIP   (CHAR_WIDTH*CHARS_PER_STRIP)


#define RX_TX_PIXEL_RATIO  (RX_CHAR_HEIGHT/TX_CHAR_HEIGHT)   // Should be an integer

#define MAX_RX_QUEUE_SIZE  10000
#define NO_CHAR   -1         //  Signifies a nonexistent xmit character

#define DIDDLE_CHAR 0x01       // Special character in the font

#define MORSE_DOT_LENGTH 14  // Number of tx data bytes that make up one morse dot 

#define NUM_XMIT_MESSAGES 9  // Number of supported canned Xmit messages



 
// Control codes sent out the Com port to the modem.
// Meaningful combinations:
//
// MODEM_RX                 // Receive
// MODEM_TX                 // Transmit silence
// MODEM_TX | MODEM_MARK    // Transmit a tone

#define MODEM_RX   0x00
#define MODEM_TX   0x04
#define MODEM_MARK 0x02



// This ought to be local to CEvmHellDoc, but then the STL template stuff chokes...
struct DataStrip {
  unsigned char pixels[COLS_PER_STRIP][RX_CHAR_HEIGHT];
};


class DrawingHint : public CObject {
public:
  int stripIndex;
  int stripCol;
};


class CEvmHellRxView;   // Forward declaration


class CEvmHellDoc : public CRichEditDoc
{
protected: // create from serialization only
	CEvmHellDoc();
	DECLARE_DYNCREATE(CEvmHellDoc)

// Attributes
public:

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CEvmHellDoc)
	public:
	virtual BOOL OnNewDocument();
	virtual void Serialize(CArchive& ar);
  virtual CRichEditCntrItem *CreateClientItem(REOBJECT *preo = NULL) const;
	virtual void OnCloseDocument();
	virtual void DeleteContents();
	virtual BOOL OnSaveDocument(LPCTSTR lpszPathName);
	virtual BOOL OnOpenDocument(LPCTSTR lpszPathName);
	protected:
	virtual BOOL SaveModified();
	//}}AFX_VIRTUAL

 
// Implementation
public:
	virtual ~CEvmHellDoc();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

// EvmHell member functions
public:
	char * getTxRxStatusStr();

  void serviceDataQueues();  // Called via periodic timer messages, etc., from App. 
  void processRxChar(int rxChar, bool updateScreen);

protected:
  void getNextTxChar(int& txChar, bool &xmitIt, bool& morse);
  int getTxFontWidth(int txChar);
  bool getTxFontBit(int txChar, int txCharCol, int txCharRow);
  void getMorseString(int txChar, CString& morseStr);
  bool activatePortThread(unsigned int priority);
  static UINT portThreadControllingProc(LPVOID param);
  UINT portThreadMainLoop();   // Called only from portThreadControllingProc()!
  void terminatePortThread();
  bool setupSerialPort(CString name);
  bool setupDummySerialPort();
  bool changeSerialPort(CString newPortName);
  void shutdownSerialPort();
  void sendTxData(int dataByte);  // Sends the given char out the serial port
  CEvmHellRxView *getRxView();    // Find the one and only Rx view.



protected:
	bool transmitting;
  bool idling;   // TRUE when transmitting, but buffer is empty
  bool stopWhenEmpty;  // TRUE when user has signaled to stop when buffer is empty
  bool justStartedTransmitting;  // TRUE when transmission has
                                // just started, but no chars have been sent

  bool processRxData;   //  TRUE when we should store and display received data.

  bool upperCaseXmit;   // If TRUE, map all xmitted chars to upper case
  bool fullDuplex;    // If TRUE, display RX data while transmitting

  bool useDiddle;   // If TRUE, send diddle char when tx buffer is empty.

  int rxCount;   // Counter of received data samples

  CString txFontFileName;   
  char txCharFonts[256][TX_FONT_WIDTH][TX_FONT_HEIGHT];

  bool carrierFlag;

  // Note:  If inlining is turned off in non-Debug mode, there seems to be some
  // problem with the queues...  So, make sure inlining is turned on.
  //
  // Above note applies to STL queues.  I replaced them with CList lists,
  // which seem to work OK.

  CList<int, int> rxDataQueue;       // Shared between threads
  CCriticalSection rxDataQueueLock;

  CList<int, int> txDataQueue;       // Shared between threads
  CCriticalSection txDataQueueLock;

  bool portThreadRunning;       // Shared between threads

  HANDLE comPort;        // Shared between threads
  bool usingDummyPort;       // Shared between threads
  CString comPortName;
  unsigned int comPortSpeed;

  CWinThread* mainThread;  // This never changes.
  CWinThread* portThread;  // This changes every time we close or open the serial port.
  bool terminatePortThreadFlag;   // Flag for signalling the thread to exit.


// Data for storing the received image strip
public:   // So view can draw it.
  vector<DataStrip*> strips;

  int nStrips;
  DataStrip *curStrip;
  int curStripIndex;
  int curStripCol;
  int curStripRow;

  bool reverseVideo;  // Set to draw black on white

  CString xmitMessageTexts[NUM_XMIT_MESSAGES];  // Canned Xmit messages.



// Generated message map functions

	//{{AFX_MSG(CEvmHellDoc)
	afx_msg void OnOperateCleartxbuffer();
	afx_msg void OnOperateTransmit();
	afx_msg void OnOperateImmediatereceive();
	afx_msg void OnOperateReceive();
	afx_msg void OnOperateDisplayoff();
	afx_msg void OnOperateDisplayon();
	afx_msg void OnUpdateOperateDisplayoff(CCmdUI* pCmdUI);
	afx_msg void OnUpdateOperateDisplayon(CCmdUI* pCmdUI);
	afx_msg void OnUpdateOperateTransmit(CCmdUI* pCmdUI);
	afx_msg void OnViewBlackonwhite();
	afx_msg void OnUpdateViewBlackonwhite(CCmdUI* pCmdUI);
	afx_msg void OnViewWhiteonblack();
	afx_msg void OnUpdateViewWhiteonblack(CCmdUI* pCmdUI);
	afx_msg void OnViewCleardisplay();
	afx_msg void OnUpdateOperateReceive(CCmdUI* pCmdUI);
	afx_msg void OnOperateUppercasexmit();
	afx_msg void OnUpdateOperateUppercasexmit(CCmdUI* pCmdUI);
	afx_msg void OnOperateFullduplex();
	afx_msg void OnUpdateOperateFullduplex(CCmdUI* pCmdUI);
	afx_msg void OnCommDummyport();
	afx_msg void OnUpdateCommDummyport(CCmdUI* pCmdUI);
 	afx_msg void OnCommCom1();
	afx_msg void OnUpdateCommCom1(CCmdUI* pCmdUI);
	afx_msg void OnCommCom2();
	afx_msg void OnUpdateCommCom2(CCmdUI* pCmdUI);
	afx_msg void OnCommCom3();
	afx_msg void OnUpdateCommCom3(CCmdUI* pCmdUI);
	afx_msg void OnOperateLoopforever();
	afx_msg void OnUpdateOperateLoopforever(CCmdUI* pCmdUI);
	afx_msg void OnOperateKick();
	afx_msg void OnEditXmitmessages();
	afx_msg void OnOperateDiddle();
	afx_msg void OnUpdateOperateDiddle(CCmdUI* pCmdUI);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_EVMHELLDOC_H__8A0E0ECF_366C_11D1_BB7C_0080489E8DF7__INCLUDED_)
