// evmhellView.h : interface of the CEvmHellRxView and CEvmHellTxView classes
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_EVMHELLVIEW_H__8A0E0ED1_366C_11D1_BB7C_0080489E8DF7__INCLUDED_)
#define AFX_EVMHELLVIEW_H__8A0E0ED1_366C_11D1_BB7C_0080489E8DF7__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "evmhellDoc.h"



#define STRIP_GAP      4     // Size in screen pixels of gap between strips
#define STRIP_HEIGHT (2*RX_CHAR_HEIGHT)
#define STRIP_PITCH   (STRIP_HEIGHT + STRIP_GAP)
#define DATA_BIT_WIDTH 3  // Width of data pixel in screen pixels
#define STRIP_WIDTH (DATA_BIT_WIDTH*(COLS_PER_STRIP+1))

#define TOP_MARGIN  4
#define BOTTOM_MARGIN  TOP_MARGIN
#define LEFT_MARGIN  4
#define RIGHT_MARGIN LEFT_MARGIN



class CEvmHellRxView : public CScrollView
{
protected: // create from serialization only
	CEvmHellRxView();
	DECLARE_DYNCREATE(CEvmHellRxView)

// Attributes
public:
	CEvmHellDoc* GetDocument();

// Operations
public:

  void ResyncScrollSizes();

  // This is called by the document after a new strip is added.
  // If the document was previously scrolled to the end, this
  // will re-scroll it to the end.
  void ScrollForNewStrip();

  void ScrollToEnd();

  // This incrementally updates the display for newly received data
  void DrawNewStripData(int strip, int col);


// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CEvmHellRxView)
	public:
	virtual void OnDraw(CDC* pDC);  // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	virtual void OnInitialUpdate();

	protected:
	virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);
	virtual void OnBeginPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnEndPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void CalcWindowRect(LPRECT lpClientRect, UINT nAdjustType = adjustBorder);
	virtual void OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint);
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~CEvmHellRxView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

  // Internally used to update strollbars, etc.
  CSize GetRxViewSize();

  void drawDataPixel(CDC *pDC, int x, int y, int val);
  void drawColumn(CDC* pDC, DataStrip *sp, int stripnum, int col, bool reverseVideo);
  CRect getStripExtent(int stripNum);

  // This structure is used to quickly draw the data columns.
  BITMAPINFO* colDIBitmap;
  char *colBitmapBits;  // The actual data

  CSize pageSize;  // Used as arguments to SetScrollSize().
  CSize lineSize;

  //CDC colBitmapDC;  // These do not seem necessary...
  //CBitmap colBitmap;


// Generated message map functions
protected:
	//{{AFX_MSG(CEvmHellRxView)
	afx_msg void OnDropFiles(HDROP hDropInfo);
	afx_msg void OnSize(UINT nType, int cx, int cy);
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

#ifndef _DEBUG  // debug version in evmhellView.cpp
inline CEvmHellDoc* CEvmHellRxView::GetDocument()
   { return (CEvmHellDoc*)m_pDocument; }
#endif



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////




class CEvmHellTxView : public CRichEditView
{
protected: // create from serialization only
	CEvmHellTxView();
	DECLARE_DYNCREATE(CEvmHellTxView)

// Attributes
public:
	CEvmHellDoc* GetDocument();

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CEvmHellTxView)
	public:
	virtual void OnDraw(CDC* pDC);  // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	virtual void OnInitialUpdate();
	virtual BOOL OnDrop(COleDataObject* pDataObject, DROPEFFECT dropEffect, CPoint point);
	virtual DROPEFFECT OnDragEnter(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point);
	virtual void OnDragLeave();
	virtual DROPEFFECT OnDragOver(COleDataObject* pDataObject, DWORD dwKeyState, CPoint point);
	protected:
	virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);
	virtual void OnBeginPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnEndPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint);
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~CEvmHellTxView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// EvmHell-specific stuff
  int prevTextLen;  // Number of chars before most recent user update.
  CHARFORMAT defaultCf;
  CHARFORMAT xmittedCf;
  CHARFORMAT unxmittedCf;   // Resets only what xmittedCf sets
  CHARFORMAT morseCf;
  int nextTxCharPos;  // Offset of next char to Tx. Increments
                       // continually from 0 as characters are sent.

  bool loopForever;    // If TRUE, send the TX data over and over.

  bool morseInputMode;   // If TRUE, the chars the user types are meant to be Morse

  //COleDropTarget  dropTarget;   // Thingie for supporting drag-and-drop into Tx window

public:
  void getNextXmitChar(int& txChar, bool& morse);
  void clearTxBuffer();

// Generated message map functions
protected:
	//{{AFX_MSG(CEvmHellTxView)
	afx_msg void OnProtected(NMHDR * pNotifyStruct, LRESULT * result);
	afx_msg void OnChange(NMHDR * pNotifyStruct, LRESULT * result);
	afx_msg void OnOperateLoopforever();
	afx_msg void OnUpdateOperateLoopforever(CCmdUI* pCmdUI);
	afx_msg void OnOperateClearxmittedchars();
	afx_msg void OnEditUnmarkxmittedchars();
	afx_msg void OnEditXmitasfeldhell();
	afx_msg void OnUpdateEditXmitasfeldhell(CCmdUI* pCmdUI);
	afx_msg void OnEditXmitasmorse();
	afx_msg void OnUpdateEditXmitasmorse(CCmdUI* pCmdUI);
	afx_msg void OnDropFiles(HDROP hDropInfo);
 	afx_msg void OnTransmitMessage();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

#ifndef _DEBUG  // debug version in evmhellView.cpp
inline CEvmHellDoc* CEvmHellTxView::GetDocument()
   { return (CEvmHellDoc*)m_pDocument; }
#endif

//////////////////////////////////////////////////////////////////////


//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_EVMHELLVIEW_H__8A0E0ED1_366C_11D1_BB7C_0080489E8DF7__INCLUDED_)
